import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

public class ThreadPoolTest {

    public static void main(String[] args) {
        /* U realnim sistemima koji opsluzuju zahtjeve korisnika (kojih moze biti veliki broj), na zahtjeve je potrebno odgovoriti u sto kracem vremenskom periodu.
        Primjeri ovih sistema su web serveri, serveri baza podataka itd.
        Idealno bi bilo svaki korisnicki zahtjev (zadatak) "dodijeliti" posebnom procesu (tredu) koji ce ga obraditi.
        Medjutim, ovaj pristup ima velike nedostatke.
        Moze se desiti da sistem vise vremena trosi na kreiranje i ubijanje ovih procesa, nego na samu obradu zahtjeva.
        Takodje, pokretanje velikog broja procesa (tredova) moze da napuni sistemsku memoriju i dovede cak i do pada servera.
        Iz ovih razloga je potrebno ograniciti broj procesa (tredova) koji odgovaraju na koristnicke zahtjeve.
        Za to se koristi ThreadPool koji predstavlja grupu worker-a koji opsluzuju zahtjeve i mogu se iznova koristiti.
        Kada dodje jedan zahtjev, bira se jedan tred iz ThreadPool-a koji ce ga obraditi.
        Nakon zavrsene obrade zahtjeva, tred se opet stavlja na raspolaganje u ThreadPool.
        Kako tredovi postoje prije nego zahtjev dodje, ne trosi se bespotrebno vrijeme na kreiranje treda.
        Ako su svi tredovi zauzeti prilikom dolaska novog zahtjeva, taj zahtjev se smjesta u red za cekanje i bice obradjen cim se neki od tredova oslobodi.
        Logika za opsluzenje zahtjeva, mora biti implementirana u klasama koje implementiraju interfejs Runnable.
         */

        ExecutorService executor = Executors.newFixedThreadPool(5);
        // Kreiramo pool od 5 tredova
        // Svi zahtjevi koji se ne mogu odmah obraditi se smjestaju u red za cekanje.
        // Ako ima previse zahtjeva i redovi za cekanje, mogu potencijalno zagusiti memoriju racunara.

        // ExecutorService executor = Executors.newCachedThreadPool();
        // Ovaj ThreadPool kreira nove tredove po potrebi (ako su svi ostali tredovi zauzeti).
        // Moze se posmatrati kao da nema red za cekanje iako se zapravo izvrsava synchronous handoff procedura.
        // Krece od 0 tredova i moze da naraste do Integer.MAX_VALUE tredova i zagusi sistem i zato treba biti oprezan kada se ovaj pool koristi.
        // Ako je tred neaktivan jedan minut, bice obrisan iz pool-a.

        // ExecutorService executor = Executors.newSingleThreadExecutor(); // pool od jednog treda

        // Kada ThreadPoolExecutor-i ne mogu kreirati nove tredove, ni smjestati nove zahtjeve u red, bacaju exception.
        // Moze se kreirati i poseban ThreadPoolExecutor() ako nijedan od ovih predefinisanih nije odgovarajuci.

        /*Moguci problemi:
        * 1. Deadlock - kao i u svim multitreding programima. Ovo je odgovornost programera.
        *               Na primjer, ako su dva zahtjeva medjusobno zavisna i jedan cega rezultate drugog,
        *               ne treba ih dodavati u isti pool, jer moze doci do deadlock-a.
        * 2. Thread Leakage - ako se neki od tredova baci Exception koji pool klasa ne hendluje, taj tred se gubi iz pool-a i nece se moci koristiti.
        * 3. Izgladnjivanje tredova (starvation) - ako je broj tredova u pool-u, veci nego sto je stvarno potrebno.
        *                                          Obicno je optimalno da broj tredova u pool-u odgovara broju procesora ili da je eventualno za jedan veci.*/
        for (int i = 0; i < 10; i++) {
            Runnable worker = new Zadatak(i);
            executor.execute(worker); // dodamo zadatak u ThreadPool
        }
        executor.shutdown(); // ova operacija se mora uraditi na kraju, inace ce tredovi ostati da se vrte u pozadini
        while(!executor.isTerminated()){} // i mora se cekati dok se svi tredovi ne ugase

        System.out.println("Kraj!");

    }
}
